#include "CellMLToRDFTree.h"
#include <nsComponentManagerUtils.h>
#include <nsServiceManagerUtils.h>
#include <nsStringAPI.h>
#include <prprf.h>

namespace CORBA
{
  class Object_ptr;
};

#include <ICORBAWrapper.h>

bool CellMLToRDFTree::mSetupStatic = false;
nsIRDFResource* CellMLToRDFTree::mRootResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mIvRootResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mTypeResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mPropResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mValueResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mValuePropResource = nsnull;
nsIRDFResource* CellMLToRDFTree::mOpenResource = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mTrueLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNumberLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mModelLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponentsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mGroupLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mGroupsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mConnectionsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mConnectionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mUnitsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mImportsLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariablesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRolesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPubIntLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPrivIntLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mBaseLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDerivedLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mPrefixLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mExponentLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mOffsetLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mMultiplierLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mUnitLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReversibleLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mIrreversibleLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReactionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariableRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReactantLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mProductLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRateLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mCatalystLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mActivatorLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mInhibitorLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mModifierLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDirectionLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mForwardLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mReverseLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mBothLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mStoichiometryLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mDeltaVariableLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mRelationshipRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNameLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mNamespaceLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponentRefLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponent1Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mComponent2Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariable1Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariable2Literal = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mVariableMappingLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mMapVariablesLiteral = nsnull;
nsIRDFLiteral* CellMLToRDFTree::mInterfaceIdLiteral[3];
nsCOMPtr<nsIRDFService> CellMLToRDFTree::mRDFService;
nsCOMPtr<nsIRDFContainerUtils> CellMLToRDFTree::mContUtils;
nsCOMPtr<nsIRDFDataSource> CellMLToRDFTree::mPersistSource;

CellMLToRDFTree::CellMLToRDFTree()
{
  mOpenSet.Init();
  mHash.Init();
}

CellMLToRDFTree::~CellMLToRDFTree()
{
}

NS_IMPL_ISUPPORTS1(CellMLToRDFTree, ICellMLToRDFTree);


void
CellMLToRDFTree::EnsureLiterals()
{
  if (mSetupStatic)
    return;
  mSetupStatic = PR_TRUE;
  mRDFService = do_GetService("@mozilla.org/rdf/rdf-service;1");
  mContUtils = do_GetService("@mozilla.org/rdf/container-utils;1");
#define RESOURCE(v,url) \
  mRDFService->GetResource(NS_LITERAL_CSTRING(url), &m##v##Resource)

  RESOURCE(Root, "http://www.cellml.org/tools/pcenv/model_tree_view");
  RESOURCE(IvRoot, "http://www.cellml.org/tools/pcenv/model_tree_view#ivonly");
  RESOURCE(Type, "http://www.cellml.org/tools/pcenv/model_tree_view#type");
  RESOURCE(Prop, "http://www.cellml.org/tools/pcenv/model_tree_view#prop");
  RESOURCE(Value, "http://www.cellml.org/tools/pcenv/model_tree_view#value");
  RESOURCE(ValueProp, "http://www.cellml.org/tools/pcenv/model_tree_view#valueprop");
  RESOURCE(Open, "http://home.netscape.com/NC-rdf#open");

#undef RESOURCE
#define LITERAL(v,val) \
  mRDFService->GetLiteral(NS_LITERAL_STRING(val).get(), &m##v##Literal)

  LITERAL(True, "true");
  LITERAL(Number, "number");
  LITERAL(Model, "Model");
  LITERAL(Components, "Components");
  LITERAL(Group, "Group");
  LITERAL(Groups, "Groups");
  LITERAL(Connections, "Connections");
  LITERAL(Connection, "Connection");
  LITERAL(Units, "Units");
  LITERAL(Imports, "Imports");
  LITERAL(Variables, "Variables");
  LITERAL(Roles, "Roles");
  LITERAL(PubInt, "Public Interface");
  LITERAL(PrivInt, "Private Interface");
  LITERAL(Base, "Base");
  LITERAL(Derived, "Derived");
  LITERAL(Prefix, "Prefix");
  LITERAL(Exponent, "Exponent");
  LITERAL(Offset, "Offset");
  LITERAL(Multiplier, "Multiplier");
  LITERAL(Unit, "Unit");
  LITERAL(Reversible, "Reversible");
  LITERAL(Irreversible, "Irreversible");
  LITERAL(Reaction, "Reaction");
  LITERAL(VariableRef, "Variable Reference");
  LITERAL(Reactant, "Reactant");
  LITERAL(Product, "Product");
  LITERAL(Rate, "Rate");
  LITERAL(Catalyst, "Catalyst");
  LITERAL(Activator, "Activator");
  LITERAL(Inhibitor, "Inhibitor");
  LITERAL(Modifier, "Modifier");
  LITERAL(Direction, "Direction");
  LITERAL(Forward, "Forward");
  LITERAL(Reverse, "Reverse");
  LITERAL(Both, "Both");
  LITERAL(Stoichiometry, "Stoichiometry");
  LITERAL(DeltaVariable, "Delta Variable");
  LITERAL(RelationshipRef, "RelationshipRef");
  LITERAL(Name, "Name");
  LITERAL(Namespace, "Namespace");
  LITERAL(ComponentRef, "ComponentRef");
  LITERAL(Component1, "Component 1");
  LITERAL(Component2, "Component 2");
  LITERAL(Variable1, "Variable 1");
  LITERAL(Variable2, "Variable 2");
  LITERAL(VariableMapping, "Variable Mapping");
  LITERAL(MapVariables, "MapVariables");
#undef LITERAL

  mRDFService->GetLiteral(NS_LITERAL_STRING("In").get(), mInterfaceIdLiteral);
  mRDFService->GetLiteral(NS_LITERAL_STRING("Out").get(),
                          mInterfaceIdLiteral + 1);
  mRDFService->GetLiteral(NS_LITERAL_STRING("None").get(),
                          mInterfaceIdLiteral + 2);

  mRDFService->GetDataSource("rdf:local-store", getter_AddRefs(mPersistSource));
}

void
CellMLToRDFTree::ExpandResourceInTree(nsIRDFResource* aResource)
{
  mPersistSource->Assert(aResource, mOpenResource, mTrueLiteral, true);
}

NS_IMETHODIMP
CellMLToRDFTree::Initialise
(
 cellml_api_IModel* aModel
)
{
  NS_ENSURE_TRUE(aModel,NS_ERROR_INVALID_ARG);
  EnsureLiterals();

  mHash.Clear();

  // Make a new datasource...
  mDatasource = do_CreateInstance
    ("@mozilla.org/rdf/datasource;1?name=in-memory-datasource");
  if (mDatasource == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;

  // Make top-level sequences...
  nsCOMPtr<nsIRDFContainer> rootseq, ivseq;
  mContUtils->MakeSeq(mDatasource, mRootResource, getter_AddRefs(rootseq));
  mContUtils->MakeSeq(mDatasource, mIvRootResource, getter_AddRefs(ivseq));

  PopulateModelIntoSequences(aModel, rootseq, ivseq);

  return NS_OK;
}

CellMLToRDFTree::ResourceAssociation::ResourceAssociation
(
 const char* aType,
 cellml_api_ICellMLElement* aElement
)
  : mType(aType), mElement(aElement)
{
}

void MakeAssociationKey(const char* aType,
                        cellml_api_ICellMLElement* aElement,
                        nsACString& aStr)
{
  nsCString tmp;
  nsCOMPtr<ICORBAWrapper> cw(do_QueryInterface(aElement));
  cw->GetObjid(getter_Copies(tmp));
  aStr.Assign(tmp);
  aStr.AppendLiteral(":");
  aStr.Append(aType);
}

already_AddRefed<nsIRDFResource>
CellMLToRDFTree::GetAssociatedResource
(
 const char* aType,
 cellml_api_ICellMLElement* aElement
)
{
  nsCOMPtr<nsIRDFResource> blank;
  mRDFService->GetAnonymousResource(getter_AddRefs(blank));

  nsCString assocKey;
  MakeAssociationKey(aType, aElement, assocKey);
  int val = 0;
  mOpenSet.Get(assocKey, &val);
  if (val == 1)
    ExpandResourceInTree(blank);

  ResourceAssociation* ra = new ResourceAssociation(aType, aElement);

  nsIRDFResource* ret = blank;
  mHash.Put(ret, ra);

  NS_ADDREF(ret);
  return ret;
}

void
CellMLToRDFTree::PopulateModelIntoSequences
(
 cellml_api_IModel* aModel,
 nsIRDFContainer* aRootseq,
 nsIRDFContainer* aIVseq
)
{
  // Create the top-level entries...
#define TOPLEVEL(x,X) \
  nsCOMPtr<nsIRDFResource> x(GetAssociatedResource("model_" #x, aModel)); \
  nsCOMPtr<nsIRDFContainer> x##Seq; \
  mDatasource->Assert(x, mTypeResource, m##X##Literal, true); \
  mDatasource->Assert(x, mPropResource, m##X##Literal, true); \
  mContUtils->MakeSeq(mDatasource, x, getter_AddRefs(x##Seq)); \
  aRootseq->AppendElement(x); \
  ExpandResourceInTree(x);

  TOPLEVEL(components,Components);
  TOPLEVEL(groups,Groups);
  TOPLEVEL(connections,Connections);
  TOPLEVEL(units,Units);
  TOPLEVEL(imports,Imports);
#undef TOPLEVEL

 {
   nsCOMPtr<cellml_api_ICellMLComponentSet> comps;
   nsCOMPtr<cellml_api_ICellMLComponentIterator> compi;
   aModel->GetLocalComponents(getter_AddRefs(comps));
   comps->IterateComponents(getter_AddRefs(compi));
   PopulateComponentsIntoSequences(compi, componentsSeq, aIVseq);
 }

 {
   nsCOMPtr<cellml_api_IGroupSet> groupss;
   nsCOMPtr<cellml_api_IGroupIterator> groupsi;
   aModel->GetGroups(getter_AddRefs(groupss));
   groupss->IterateGroups(getter_AddRefs(groupsi));
   PopulateGroupsIntoSequences(groupsi, groupsSeq);
 }

 {
   nsCOMPtr<cellml_api_IConnectionSet> conns;
   nsCOMPtr<cellml_api_IConnectionIterator> conni;
   aModel->GetConnections(getter_AddRefs(conns));
   conns->IterateConnections(getter_AddRefs(conni));
   PopulateConnectionsIntoSequences(conni, connectionsSeq);
 }

 {
   nsCOMPtr<cellml_api_IUnitsSet> unitss;
   nsCOMPtr<cellml_api_IUnitsIterator> unitsi;
   aModel->GetLocalUnits(getter_AddRefs(unitss));
   unitss->IterateUnits(getter_AddRefs(unitsi));
   PopulateUnitsIntoSequences(unitsi, unitsSeq);
 }
 
 {
   nsCOMPtr<cellml_api_ICellMLImportSet> importss;
   nsCOMPtr<cellml_api_ICellMLImportIterator> importsi;
   aModel->GetImports(getter_AddRefs(importss));
   importss->IterateImports(getter_AddRefs(importsi));
   PopulateImportsIntoSequences(importsi, importsSeq, aIVseq);
 }
}

void
CellMLToRDFTree::PopulateComponentsIntoSequences
(
 cellml_api_ICellMLComponentIterator* aCompIt,
 nsIRDFContainer* aCompSeq,
 nsIRDFContainer* aIVseq
)
{
  nsCOMPtr<cellml_api_ICellMLComponent> comp;
  while (PR_TRUE)
  {
    aCompIt->NextComponent(getter_AddRefs(comp));
    if (comp == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("component", comp));

    nsCOMPtr<nsIRDFLiteral> nameLit;
    nsString compNameStr;
    comp->GetName(getter_Copies(compNameStr));
    mRDFService->GetLiteral(compNameStr.get(), getter_AddRefs(nameLit));

    mDatasource->Assert(resource, mTypeResource, nameLit, true);
    mDatasource->Assert(resource, mPropResource, mComponentsLiteral, true);

    // Make a sequence...
    nsCOMPtr<nsIRDFContainer> compSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(compSeq));

    nsCOMPtr<cellml_api_ICellMLVariableSet> vars;
    comp->GetVariables(getter_AddRefs(vars));
    nsCOMPtr<cellml_api_ICellMLVariableIterator> vari;
    vars->IterateVariables(getter_AddRefs(vari));
    PopulateVariablesIntoSequences(vari, compSeq, aIVseq);

    nsCOMPtr<cellml_api_IUnitsSet> units;
    comp->GetUnits(getter_AddRefs(units));
    nsCOMPtr<cellml_api_IUnitsIterator> unitsi;
    units->IterateUnits(getter_AddRefs(unitsi));
    PopulateUnitsIntoSequences(unitsi, compSeq);

    nsCOMPtr<cellml_api_IReactionSet> reactions;
    comp->GetReactions(getter_AddRefs(reactions));
    nsCOMPtr<cellml_api_IReactionIterator> reacti;
    reactions->IterateReactions(getter_AddRefs(reacti));
    PopulateReactionsIntoSequences(reacti, compSeq);

    aCompSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateGroupsIntoSequences
(
 cellml_api_IGroupIterator* aGroupIt,
 nsIRDFContainer* aGroupSeq
)
{
  nsCOMPtr<cellml_api_IGroup> group;
  while (PR_TRUE)
  {
    aGroupIt->NextGroup(getter_AddRefs(group));

    if (group == nsnull)
      break;

    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("group", group));

    mDatasource->Assert(resource, mTypeResource, mGroupLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mGroupsLiteral, PR_TRUE);
    
    // Make a sequence...
    nsCOMPtr<nsIRDFContainer> groupsSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(groupsSeq));
    
    nsCOMPtr<cellml_api_IRelationshipRefSet> rrs;
    group->GetRelationshipRefs(getter_AddRefs(rrs));
    nsCOMPtr<cellml_api_IRelationshipRefIterator> rri;
    rrs->IterateRelationshipRefs(getter_AddRefs(rri));
    PopulateRelationshipRefsIntoSequences(rri, groupsSeq);
    
    nsCOMPtr<cellml_api_IComponentRefSet> crs;
    group->GetComponentRefs(getter_AddRefs(crs));
    nsCOMPtr<cellml_api_IComponentRefIterator> cri;
    crs->IterateComponentRefs(getter_AddRefs(cri));
    PopulateComponentRefsIntoSequences(cri, groupsSeq);

    aGroupSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateConnectionsIntoSequences
(
 cellml_api_IConnectionIterator* aConnIt,
 nsIRDFContainer* aConnSeq
)
{
  nsCOMPtr<cellml_api_IConnection> conn;
  while (PR_TRUE)
  {
    aConnIt->NextConnection(getter_AddRefs(conn));
    if (conn == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("connection", conn));

    mDatasource->Assert(resource, mTypeResource, mConnectionLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mConnectionsLiteral, PR_TRUE);

    nsCOMPtr<nsIRDFContainer> connSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(connSeq));

    nsCOMPtr<cellml_api_IMapComponents> mapcom;
    conn->GetComponentMapping(getter_AddRefs(mapcom));

    nsCOMPtr<nsIRDFResource> mapCompRes
      (GetAssociatedResource("connection_comp1", conn));
    mDatasource->Assert(mapCompRes, mTypeResource, mComponent1Literal,
                        PR_TRUE);
    nsCOMPtr<nsIRDFLiteral> lit;
    nsString str;
    mapcom->GetFirstComponentName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(mapCompRes, mValueResource, lit, PR_TRUE);
    connSeq->AppendElement(mapCompRes);

    mapCompRes = GetAssociatedResource("connection_comp2", conn);
    mDatasource->Assert(mapCompRes, mTypeResource, mComponent2Literal,
                        PR_TRUE);
    mapcom->GetSecondComponentName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(mapCompRes, mValueResource, lit, PR_TRUE);
    connSeq->AppendElement(mapCompRes);

    nsCOMPtr<cellml_api_IMapVariablesSet> mvs;
    conn->GetVariableMappings(getter_AddRefs(mvs));
    nsCOMPtr<cellml_api_IMapVariablesIterator> mvi;
    mvs->IterateMapVariables(getter_AddRefs(mvi));
    
    PopulateMapVariablesIntoSequences(mvi, connSeq);

    aConnSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateUnitsIntoSequences
(
 cellml_api_IUnitsIterator* aUnitsIt,
 nsIRDFContainer* aUnitsSeq
)
{
  nsCOMPtr<cellml_api_IUnits> units;
  while (PR_TRUE)
  {
    aUnitsIt->NextUnits(getter_AddRefs(units));
    if (units == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("units", units));

    nsString str;
    units->GetName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitsLiteral, PR_TRUE);
    PRBool isBase;
    units->GetIsBaseUnits(&isBase);
    mDatasource->Assert(resource, mValueResource, isBase ? mBaseLiteral :
                        mDerivedLiteral, PR_TRUE);

    nsCOMPtr<nsIRDFContainer> unitsSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(unitsSeq));

    nsCOMPtr<cellml_api_IUnitSet> uns;
    units->GetUnitCollection(getter_AddRefs(uns));
    nsCOMPtr<cellml_api_IUnitIterator> uni;
    uns->IterateUnits(getter_AddRefs(uni));
    PopulateUnitIntoSequences(uni, unitsSeq);

    aUnitsSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateImportsIntoSequences
(
 cellml_api_ICellMLImportIterator* aImpIt,
 nsIRDFContainer* aImpSeq,
 nsIRDFContainer* aIVseq
)
{
  nsCOMPtr<cellml_api_ICellMLImport> import;
  while (PR_TRUE)
  {
    aImpIt->NextImport(getter_AddRefs(import));
    if (import == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import", import));

    nsString str;
    nsCOMPtr<cellml_api_IURI> xlink;
    import->GetXlinkHref(getter_AddRefs(xlink));
    xlink->GetAsText(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mImportsLiteral, PR_TRUE);

    nsCOMPtr<nsIRDFContainer> importSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(importSeq));

    nsCOMPtr<nsIRDFResource> icresource
      (GetAssociatedResource("import_components", import));
    mDatasource->Assert(icresource, mTypeResource, mComponentsLiteral,
                        PR_TRUE);
    mDatasource->Assert(icresource, mPropResource, mComponentsLiteral,
                        PR_TRUE);
    nsCOMPtr<nsIRDFContainer> impCompSeq;
    mContUtils->MakeSeq(mDatasource, icresource, getter_AddRefs(impCompSeq));   
    importSeq->AppendElement(icresource);

    nsCOMPtr<nsIRDFResource> iuresource
      (GetAssociatedResource("import_units", import));
    mDatasource->Assert(iuresource, mTypeResource, mUnitsLiteral,
                        PR_TRUE);
    mDatasource->Assert(iuresource, mPropResource, mUnitsLiteral,
                        PR_TRUE);
    nsCOMPtr<nsIRDFContainer> impUnitsSeq;
    mContUtils->MakeSeq(mDatasource, iuresource, getter_AddRefs(impUnitsSeq));
    importSeq->AppendElement(iuresource);
    
    nsCOMPtr<cellml_api_IImportComponentSet> ics;
    import->GetComponents(getter_AddRefs(ics));
    nsCOMPtr<cellml_api_IImportComponentIterator> ici;
    ics->IterateImportComponents(getter_AddRefs(ici));
    PopulateImportComponentsIntoSequences(ici, impCompSeq);

    nsCOMPtr<cellml_api_IImportUnitsSet> ius;
    import->GetUnits(getter_AddRefs(ius));
    nsCOMPtr<cellml_api_IImportUnitsIterator> iui;
    ius->IterateImportUnits(getter_AddRefs(iui));
    PopulateImportUnitsIntoSequences(iui, impUnitsSeq);

    nsCOMPtr<cellml_api_IModel> mod;
    import->GetImportedModel(getter_AddRefs(mod));
    nsCOMPtr<nsIRDFResource> modresource
      (GetAssociatedResource("import_model", mod));
    mDatasource->Assert(modresource, mTypeResource, mModelLiteral, PR_TRUE);
    mDatasource->Assert(modresource, mPropResource, mModelLiteral, PR_TRUE);
    importSeq->AppendElement(modresource);
    nsCOMPtr<nsIRDFContainer> modSeq;
    mContUtils->MakeSeq(mDatasource, modresource, getter_AddRefs(modSeq));

    PopulateModelIntoSequences(mod, modSeq, aIVseq);

    aImpSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateVariablesIntoSequences
(
 cellml_api_ICellMLVariableIterator* aVarIt,
 nsIRDFContainer* aVarSeq,
 nsIRDFContainer* aIVseq
)
{
  nsCOMPtr<cellml_api_ICellMLVariable> var;
  while (PR_TRUE)
  {
    aVarIt->NextVariable(getter_AddRefs(var));
    if (var == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("variable", var));

    nsCOMPtr<nsIRDFContainer> varSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(varSeq));

    nsString str;
    var->GetName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mVariablesLiteral, PR_TRUE);
    
    nsString iv;
    nsresult ret;
    ret = var->GetInitialValue(getter_Copies(iv));
    if (NS_SUCCEEDED(ret) && !iv.Equals(NS_LITERAL_STRING("")))
    {
      mRDFService->GetLiteral(iv.get(), getter_AddRefs(lit));
      mDatasource->Assert(resource, mValueResource, lit, PR_TRUE);
      mDatasource->Assert(resource, mValuePropResource, mNumberLiteral,
                          PR_TRUE);
      // If the variable has an initial value, put it into the IV seq too...
      aIVseq->AppendElement(resource);
    }

    nsCOMPtr<nsIRDFResource> pubIntRes
      (GetAssociatedResource("public_interface", var));
    mDatasource->Assert(pubIntRes, mTypeResource, mPubIntLiteral,
                        PR_TRUE);
    PRUint32 pubint;
    var->GetPublicInterface(&pubint);
    mDatasource->Assert(pubIntRes, mValueResource, mInterfaceIdLiteral[pubint],
                        PR_TRUE);
    varSeq->AppendElement(pubIntRes);

    nsCOMPtr<nsIRDFResource> privIntRes
      (GetAssociatedResource("private_interface", var));
    mDatasource->Assert(privIntRes, mTypeResource, mPrivIntLiteral,
                        PR_TRUE);
    PRUint32 privint;
    var->GetPrivateInterface(&privint);
    mDatasource->Assert(privIntRes, mValueResource, mInterfaceIdLiteral[privint],
                        PR_TRUE);
    varSeq->AppendElement(privIntRes);

    nsCOMPtr<nsIRDFResource> unitsRes
      (GetAssociatedResource("variable_units", var));
    mDatasource->Assert(unitsRes, mTypeResource, mUnitsLiteral, PR_TRUE);
    var->GetUnitsName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(unitsRes, mValueResource, lit, PR_TRUE);
    aVarSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateReactionsIntoSequences
(
 cellml_api_IReactionIterator* aReactIt,
 nsIRDFContainer* aReactSeq
)
{
  nsCOMPtr<cellml_api_IReaction> react;
  while (PR_TRUE)
  {
    aReactIt->NextReaction(getter_AddRefs(react));
    if (react == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("reaction", react));
    nsCOMPtr<nsIRDFContainer> reactSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(reactSeq));

    PRBool isReversible;
    react->GetReversible(&isReversible);
    mDatasource->Assert(resource, mTypeResource,
                        isReversible ? mReversibleLiteral :
                        mIrreversibleLiteral, PR_TRUE);
    mDatasource->Assert(resource, mPropResource,
                        mReactionLiteral, PR_TRUE);

    nsCOMPtr<cellml_api_IVariableRefSet> vrs;
    react->GetVariableReferences(getter_AddRefs(vrs));
    nsCOMPtr<cellml_api_IVariableRefIterator> vri;
    vrs->IterateVariableRefs(getter_AddRefs(vri));
    PopulateVariableRefsIntoSequences(vri, reactSeq);

    aReactSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateRelationshipRefsIntoSequences
(
 cellml_api_IRelationshipRefIterator* aRRIt,
 nsIRDFContainer* aRRSeq
)
{
  nsCOMPtr<cellml_api_IRelationshipRef> rr;
  while (PR_TRUE)
  {
    aRRIt->NextRelationshipRef(getter_AddRefs(rr));
    if (rr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("relationshipRef", rr));
    nsCOMPtr<nsIRDFContainer> rrSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(rrSeq));

    nsString str;
    rr->GetRelationship(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mRelationshipRefLiteral,
                        PR_TRUE);

    nsCOMPtr<nsIRDFResource> namersrc
      (GetAssociatedResource("relationshipRef_name", rr));
    mDatasource->Assert(namersrc, mTypeResource, mNameLiteral, PR_TRUE);
    rr->GetName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(namersrc, mValueResource, lit, PR_TRUE);
    rrSeq->AppendElement(namersrc);

    nsCOMPtr<nsIRDFResource> nsrsrc
      (GetAssociatedResource("relationshipRef_namespace", rr));
    mDatasource->Assert(nsrsrc, mTypeResource, mNamespaceLiteral, PR_TRUE);
    rr->GetRelationshipNamespace(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(nsrsrc, mValueResource, lit, PR_TRUE);
    rrSeq->AppendElement(nsrsrc);

    aRRSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateComponentRefsIntoSequences
(
 cellml_api_IComponentRefIterator* aCRIt,
 nsIRDFContainer* aCRSeq
)
{
  nsCOMPtr<cellml_api_IComponentRef> cr;
  while (PR_TRUE)
  {
    aCRIt->NextComponentRef(getter_AddRefs(cr));
    if (cr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("component_ref", cr));
    nsCOMPtr<nsIRDFContainer> crSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(crSeq));
    
    nsString str;
    cr->GetComponentName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mComponentRefLiteral,
                        PR_TRUE);

    nsCOMPtr<cellml_api_IComponentRefSet> crs;
    cr->GetComponentRefs(getter_AddRefs(crs));
    nsCOMPtr<cellml_api_IComponentRefIterator> cri;
    crs->IterateComponentRefs(getter_AddRefs(cri));
    PopulateComponentRefsIntoSequences(cri, crSeq);

    aCRSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateMapVariablesIntoSequences
(
 cellml_api_IMapVariablesIterator* aMVIt,
 nsIRDFContainer* aMVSeq
)
{
  nsCOMPtr<cellml_api_IMapVariables> mv;
  while (PR_TRUE)
  {
    aMVIt->NextMapVariable(getter_AddRefs(mv));
    if (mv == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("map_variable", mv));

    nsCOMPtr<nsIRDFContainer> mvSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(mvSeq));

    mDatasource->Assert(resource, mTypeResource, mVariableMappingLiteral,
                        PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mMapVariablesLiteral,
                        PR_TRUE);

    nsCOMPtr<nsIRDFResource> v1rsrc
      (GetAssociatedResource("map_variable_var1", mv));
    mDatasource->Assert(v1rsrc, mTypeResource, mVariable1Literal, PR_TRUE);
    nsString str;
    mv->GetFirstVariableName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(v1rsrc, mValueResource, lit, PR_TRUE);
    mvSeq->AppendElement(v1rsrc);

    nsCOMPtr<nsIRDFResource> v2rsrc
      (GetAssociatedResource("map_variable_var2", mv));
    mDatasource->Assert(v2rsrc, mTypeResource, mVariable2Literal, PR_TRUE);
    mv->GetSecondVariableName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(v2rsrc, mValueResource, lit, PR_TRUE);
    mvSeq->AppendElement(v2rsrc);

    aMVSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateUnitIntoSequences
(
 cellml_api_IUnitIterator* aUnIt,
 nsIRDFContainer* aUnitsSeq
)
{
  nsCOMPtr<cellml_api_IUnit> un;
  while (PR_TRUE)
  {
    aUnIt->NextUnit(getter_AddRefs(un));
    if (un == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("unit", un));

    nsCOMPtr<nsIRDFContainer> unSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(unSeq));

    nsString str;
    un->GetUnits(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitLiteral, PR_TRUE);

    nsCOMPtr<nsIRDFResource> prefixResource
      (GetAssociatedResource("unit_prefix", un));
    mDatasource->Assert(prefixResource, mTypeResource, mPrefixLiteral,
                        PR_TRUE);
    PRInt32 value;
    un->GetPrefix(&value);
    nsCOMPtr<nsIRDFInt> ilit;
    mRDFService->GetIntLiteral(value, getter_AddRefs(ilit));
    mDatasource->Assert(prefixResource, mValueResource, ilit, PR_TRUE);
    unSeq->AppendElement(prefixResource);

    nsCOMPtr<nsIRDFResource> mupResource
      (GetAssociatedResource("unit_multiplier", un));
    mDatasource->Assert(mupResource, mTypeResource, mPrefixLiteral,
                        PR_TRUE);
    double dvalue;
    char buf[40];
    un->GetMultiplier(&dvalue);
    PR_snprintf(buf, 40, "%f", dvalue);
    NS_ConvertUTF8toUTF16 wbuf(buf);
    mRDFService->GetLiteral(wbuf.get(), getter_AddRefs(lit));
    mDatasource->Assert(prefixResource, mValueResource, lit, PR_TRUE);
    unSeq->AppendElement(prefixResource);

    nsCOMPtr<nsIRDFResource> offsResource
      (GetAssociatedResource("unit_offset", un));
    mDatasource->Assert(offsResource, mTypeResource, mOffsetLiteral,
                        PR_TRUE);
    un->GetOffset(&dvalue);
    PR_snprintf(buf, 40, "%f", dvalue);
    NS_ConvertUTF8toUTF16 wbufo(buf);
    mRDFService->GetLiteral(wbufo.get(), getter_AddRefs(lit));
    mDatasource->Assert(offsResource, mValueResource, lit, PR_TRUE);
    unSeq->AppendElement(offsResource);

    nsCOMPtr<nsIRDFResource> expResource
      (GetAssociatedResource("unit_exponent", un));
    mDatasource->Assert(expResource, mTypeResource, mExponentLiteral,
                        PR_TRUE);
    un->GetExponent(&dvalue);
    PR_snprintf(buf, 40, "%f", dvalue);
    NS_ConvertUTF8toUTF16 wbufe(buf);
    mRDFService->GetLiteral(wbufe.get(), getter_AddRefs(lit));
    mDatasource->Assert(expResource, mValueResource, lit, PR_TRUE);
    unSeq->AppendElement(expResource);

    aUnitsSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateImportComponentsIntoSequences
(
 cellml_api_IImportComponentIterator* aICIt,
 nsIRDFContainer* aICSeq
)
{
  nsCOMPtr<cellml_api_IImportComponent> ic;
  while (PR_TRUE)
  {
    aICIt->NextImportComponent(getter_AddRefs(ic));
    if (ic == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import_component", ic));

    nsString str;
    ic->GetName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mComponentsLiteral, PR_TRUE);

    aICSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateImportUnitsIntoSequences
(
 cellml_api_IImportUnitsIterator* aIUIt,
 nsIRDFContainer* aIUSeq
)
{
  nsCOMPtr<cellml_api_IImportUnits> iu;
  while (PR_TRUE)
  {
    aIUIt->NextImportUnits(getter_AddRefs(iu));
    if (iu == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("import_unit", iu));

    nsString str;
    iu->GetName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mUnitsLiteral, PR_TRUE);

    aIUSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateVariableRefsIntoSequences
(
 cellml_api_IVariableRefIterator* aVRIt,
 nsIRDFContainer* aVRSeq
)
{
  nsCOMPtr<cellml_api_IVariableRef> vr;
  while (PR_TRUE)
  {
    aVRIt->NextVariableRef(getter_AddRefs(vr));
    if (vr == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("variable_ref", vr));

    nsCOMPtr<nsIRDFContainer> vrSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(vrSeq));

    nsString str;
    vr->GetVariableName(getter_Copies(str));
    nsCOMPtr<nsIRDFLiteral> lit;
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mVariableRefLiteral,
                        PR_TRUE);

    nsCOMPtr<cellml_api_IRoleSet> roles;
    vr->GetRoles(getter_AddRefs(roles));
    nsCOMPtr<cellml_api_IRoleIterator> rolei;
    roles->IterateRoles(getter_AddRefs(rolei));
    PopulateRolesIntoSequences(rolei, vrSeq);

    aVRSeq->AppendElement(resource);
  }
}

void
CellMLToRDFTree::PopulateRolesIntoSequences
(
 cellml_api_IRoleIterator* aRoleIt,
 nsIRDFContainer* aRoleSeq
)
{
  nsCOMPtr<cellml_api_IRole> role;
  while (PR_TRUE)
  {
    aRoleIt->NextRole(getter_AddRefs(role));
    if (role == nsnull)
      break;
    nsCOMPtr<nsIRDFResource> resource
      (GetAssociatedResource("role", role));

    nsCOMPtr<nsIRDFContainer> roleSeq;
    mContUtils->MakeSeq(mDatasource, resource, getter_AddRefs(roleSeq));

    PRUint32 roletype;
    role->GetVariableRole(&roletype);
    nsCOMPtr<nsIRDFLiteral> lit;
    switch (roletype)
    {
    case cellml_api_IRole::REACTANT:
      lit = mReactantLiteral;
      break;
    case cellml_api_IRole::PRODUCT:
      lit = mProductLiteral;
      break;
    case cellml_api_IRole::RATE:
      lit = mRateLiteral;
      break;
    case cellml_api_IRole::CATALYST:
      lit = mCatalystLiteral;
      break;
    case cellml_api_IRole::ACTIVATOR:
      lit = mActivatorLiteral;
      break;
    case cellml_api_IRole::INHIBITOR:
      lit = mInhibitorLiteral;
      break;
    case cellml_api_IRole::MODIFIER:
      lit = mModifierLiteral;
      break;
    }

    mDatasource->Assert(resource, mTypeResource, lit, PR_TRUE);
    mDatasource->Assert(resource, mPropResource, mRolesLiteral,
                        PR_TRUE);

    nsCOMPtr<nsIRDFResource> dirrsrc
      (GetAssociatedResource("role_direction", role));

    PRUint32 dirtype;
    role->GetDirection(&dirtype);
    switch (dirtype)
    {
    case cellml_api_IRole::FORWARD:
      lit = mForwardLiteral;
      break;
    case cellml_api_IRole::REVERSE:
      lit = mReverseLiteral;
      break;
    case cellml_api_IRole::BOTH:
      lit = mBothLiteral;
      break;
    }

    mDatasource->Assert(dirrsrc, mTypeResource, mDirectionLiteral, PR_TRUE);
    mDatasource->Assert(dirrsrc, mValueResource, lit, PR_TRUE);
    roleSeq->AppendElement(dirrsrc);

    nsCOMPtr<nsIRDFResource> stoichrsrc
      (GetAssociatedResource("role_stoich", role));
    mDatasource->Assert(stoichrsrc, mTypeResource, mStoichiometryLiteral,
                        PR_TRUE);
    double dvalue;
    char buf[40];
    role->GetStoichiometry(&dvalue);
    PR_snprintf(buf, 40, "%f", dvalue);
    NS_ConvertUTF8toUTF16 wbuf(buf);
    mRDFService->GetLiteral(wbuf.get(), getter_AddRefs(lit));
    mDatasource->Assert(stoichrsrc, mValueResource, lit, PR_TRUE);
    roleSeq->AppendElement(stoichrsrc);

    nsCOMPtr<nsIRDFResource> dvrsrc
      (GetAssociatedResource("role_deltavariable", role));
    mDatasource->Assert(dvrsrc, mTypeResource, mDeltaVariableLiteral,
                        PR_TRUE);
    nsString str;
    role->GetDeltaVariableName(getter_Copies(str));
    mRDFService->GetLiteral(str.get(), getter_AddRefs(lit));
    mDatasource->Assert(dvrsrc, mValueResource, lit, PR_TRUE);
    roleSeq->AppendElement(dvrsrc);
    
    aRoleSeq->AppendElement(resource);
  }
}

PLDHashOperator
CellMLToRDFTree::rememberExpansionCallback
(
 const void* aKey,
 CellMLToRDFTree::ResourceAssociation* aRa,
 void* aInst
)
{
  NS_REINTERPRET_CAST(CellMLToRDFTree*, aInst)->RememberExpansion
    (
     NS_REINTERPRET_CAST(const nsIRDFResource*, aKey),
     aRa
    );
  return PL_DHASH_NEXT;
}

NS_IMETHODIMP
CellMLToRDFTree::RememberExpansions()
{
  mOpenSet.Clear();
  mHash.EnumerateRead(rememberExpansionCallback, this);
  return NS_OK;
}

void
CellMLToRDFTree::RememberExpansion(const nsIRDFResource* aRsrc,
                                   ResourceAssociation* aRa)
{
  nsCString assocKey;
  MakeAssociationKey(aRa->mType, aRa->mElement, assocKey);
  PRBool ha = PR_FALSE;
  mPersistSource->HasAssertion(const_cast<nsIRDFResource*>(aRsrc),
                               mOpenResource, mTrueLiteral, true, &ha);
  if (ha)
    mOpenSet.Put(assocKey, 1);
}

NS_IMETHODIMP
CellMLToRDFTree::GetDatasource
(
 nsIRDFDataSource** aDatasource
)
{
  *aDatasource = mDatasource;
  if (*aDatasource != nsnull)
    NS_ADDREF(*aDatasource);
  return NS_OK;
}

NS_IMETHODIMP
CellMLToRDFTree::GetInformationForResource
(
 nsIRDFResource* aResource,
 cellml_api_ICellMLElement** aAssociatedElement,
 nsACString& aType
)
{
  ResourceAssociation* ra;
  mHash.Get(aResource, &ra);
  if (ra == nsnull)
    return NS_ERROR_FAILURE;
  aType.Assign(ra->mType);
  *aAssociatedElement = ra->mElement;
  if (*aAssociatedElement != nsnull)
    NS_ADDREF(*aAssociatedElement);
  return NS_OK;
}
